package tp4509.picross;

import static dm4509.common.Log.COMM;


import static dm4509.common.Log.LOG_ON;



import java.io.IOException;
import java.nio.channels.SocketChannel;

import org.apache.log4j.Level;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import tp4509.tcp.MessageProtocole;
import tp4509.tcp.Protocole;

/**
 * Classe donnant accès aux informations utiles pour l'interface graphique.
 * @author Eric Lallet.
 *
 */
public class PicrossApi   {
	/**
	 * 
	 */
	public static final int SANSINDICE = -1;

	/**
	 * Les grilles de picross si la construction du picross est locale.
	 */
	private Picross picross = null;

	/**
	 * La connexion au serveur si la construction du picross a été faite par un serveur distant.
	 */
	private SocketChannel rwChan = null;
	/**
	 * bloque la consultations des attributs de tailles tant qu'ils ne sont pas initialisés.
	 */
	private boolean initialisationFinie = false; 
	/**
	 * hauteur de la grille de jeu.
	 */

	private int hauteur;
	/**
	 * largeur de la grille de jeu.
	 */
	private int largeur;
	/**
	 * hauteur de la zone des indices sur les colonnes.
	 */
	private int hauteurIndices;
	/**
	 * largeur de la zone des indices sur les lignes.
	 */
	private int largeurIndices;

	/**
	 * la grille de jeu.
	 */
	private int[][] grille;
	/**
	 * les indices sur les lignes.
	 */
	private int[][] indicesLigne;
	/**
	 * les indices sur les colonnes.
	 */
	private int[][] indicesColonne;
	/**
	 * statut qui indique si la partie est gagnée.
	 */
	private boolean gagne = false;
	/**
	 * statut qui indique si la connexion est perdue.
	 */
	private boolean connexionPerdue = false;

	/**
	 * date du dernier changement des valeurs.
	 * Sert de propriété surveillée par l'interface graphique pour savoir quand il faut
	 * actualiser les données affichées.
	 */
	private Long lastChange = 0L;

	/**
	 * Gestionnaire des changements des propritées.
	 * L'api previent des changement des données avec un PropertyChangeListener.
	 * La proprité surveillée est la valeur de lastChange.
	 */
	private PropertyChangeSupport changeSupport = new PropertyChangeSupport(this);

	/**
	 * Constructeur pour jeu en local.
	 * @param fileName
	 *        nom du fichier contenant le picross.
	 * @throws FileFormatException 
	 *        exception pour des erreurs dans le fichier picross.
	 * @throws IOException 
	 *        toutes les exceptions d'entrées/sorties.
	 */
	public PicrossApi(final String fileName)  throws IOException, FileFormatException  {
		picross = new Picross(fileName);
		hauteur = picross.getHauteur();
		largeur = picross.getLargeur();
		hauteurIndices = picross.getHauteurIndices();
		largeurIndices = picross.getLargeurIndices();
		grille = new int[largeur][hauteur];

		indicesLigne = new int[largeurIndices][hauteur];
		for (int col = 0; col < largeurIndices; col++) {
			for (int lig = 0;  lig < hauteur; lig++) {
				indicesLigne[col][lig] = picross.getIndiceLigne(col, lig);
			}
		}
		indicesColonne = new int[largeur][hauteurIndices];
		for (int lig = 0; lig < hauteurIndices; lig++) {
			for (int col = 0;  col < largeur; col++) {
				indicesColonne[col][lig] = picross.getIndiceColonne(col, lig);
			}
		}
		initialisationFinie = true;

	}

	/**
	 * Contructeur pour jeu en réseau.
	 * @param rwChan
	 *        canal de la connexion reliée au serveur.
	 */
	public PicrossApi(final SocketChannel rwChan) {
		this.rwChan = rwChan;
	}

	/**
	 * Fixe la taille de la grille de jeu. Est appelée par le client lors de la réception
	 * du message TAILLE_GRILLE.
	 * @param largeur
	 *        largeur de la grille de jeu.
	 * @param hauteur
	 *        hauteur de la grille de jeu.
	 */
	public void setTailleGrille(final int largeur, final int hauteur) {
		this.setLastChange(System.currentTimeMillis());
		if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
			COMM.trace("[PicrossApi::setTailleGrille] largeur = " + largeur + ", hauteur = " + hauteur);
		}
		this.largeur = largeur;
		this.hauteur = hauteur;
		grille = new int[largeur][hauteur];
		for (int col = 0; col < largeur; col++) {
			for (int lig = 0; lig < hauteur; lig++) {
				grille[col][lig] = PicrossColor.NONE.ordinal();
			}
		}
	}

	/**
	 * Fixe la taille des grilles des indices. Est appelée par le client lors de la réception
	 * du message TAILLE_INDICES.
	 * 
	 * @param largeurIndice
	 *        largeur de la grille des indices sur les lignes.
	 * @param hauteurIndice
	 *        hauetur de la grille des indices sur les colones.
	 */
	public void setTailleIndices(final int largeurIndice, final int hauteurIndice) {
		this.setLastChange(System.currentTimeMillis());
		if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
			COMM.trace("[PicrossApi::setTailleIndices] largeur = " + largeur + ", hauteur = " + hauteur);
		}
		this.largeurIndices = largeurIndice;
		this.hauteurIndices = hauteurIndice;
		indicesColonne = new int[largeur][hauteurIndices];
		for (int col = 0; col < largeur; col++) {
			for (int lig = 0; lig < hauteurIndices; lig++) {
				indicesColonne[col][lig] = SANSINDICE;
			}
		}
		indicesLigne = new int[largeurIndices][hauteur];
		for (int col = 0; col < largeurIndices; col++) {
			for (int lig = 0; lig < hauteur; lig++) {
				indicesLigne[col][lig] = SANSINDICE;
			}
		}

	}


	/**
	 * Fixe la valeur d'une case de la grille de jeu. Est appelé par l'interface graphique
	 * lorsque le joueur clique sur une case et par le serveur lors de la réception
	 * du message VALEUR_GRILLE_CLIENT. 
	 * @param col
	 *           colonne de la grille de jeu.
	 * @param lig
	 *           ligne de la grille de jeu.
	 * @param val
	 *           valeur à placer dans la case.
	 */
	public void fixerCase(final int col, final int lig, final int val)  {
		this.setLastChange(System.currentTimeMillis());
		grille[col][lig] = val;
		if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
			COMM.trace("[PicrossApi::changerCase] col = " + col + ", lig = " + lig);
		}
		if (picross != null) {
			picross.fixerCase(col, lig, val);
		} else {
			try {
				MessageProtocole message = new MessageProtocole(Protocole.VALEUR_GRILLE_CLIENT, col, lig, val);
				rwChan.write(message.getBuffer());
			} catch (IOException e) {
				connexionPerdue = true;
			}
		}
	}
	/**
	 * Fixe la valeur d'une case de la grille de jeu. Est appelé par  le client lors de la réception
	 * du message VALEUR_GRILLE_SERVEUR.
	 * @param col
	 *           colonne de la grille de jeu.
	 * @param lig
	 *           ligne de la grille de jeu.
	 * @param val
	 *           valeur à placer dans la case.
	 */
	public void changerCaseServeur(final int col, final int lig, final int val)  {
		this.setLastChange(System.currentTimeMillis());
		if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
			COMM.trace("[PicrossApi::changerCaseServeur] col = " + col + ", lig = " + lig  + " val =" + val);
		}
		grille[col][lig] = val;
	}

	/**
	 * Fixe la valeur d'un indice sur les lignes. Est appelé par  le client lors de la réception
	 * du message INDICE_LIGNE.
	 * @param col
	 *           colonne de la case dans la grille des indices sur les lignes.
	 * @param lig
	 *           ligne de la case dans la grille des indices sur les lignes.
	 * @param val
	 *           la valeur de l'indice.
	 */
	public void changerIndiceLigne(final int col, final int lig, final int val) {
		this.setLastChange(System.currentTimeMillis());
		if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
			COMM.trace("[PicrossApi::changerIndiceLigne] col = " + col + ", lig = " + lig  + " val =" + val);
		}

		indicesLigne[col][lig] = val;
	}

	/**
	 * Fixe la valeur d'un indice sur les colonne. Est appelé par  le client lors de la réception
	 * du message INDICE_COLONNE.
	 * @param col
	 *           colonne de la case dans la grille des indices sur les colonnes.
	 * @param lig
	 *           ligne de la case dans la grille des indices sur les colonnes.
	 * @param val
	 *           la valeur de l'indice.
	 */
	public void changerIndiceColonne(final int col, final int lig, final int val) {
		this.setLastChange(System.currentTimeMillis());
		if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
			COMM.trace("[PicrossApi::changerIndiceColonne] col = " + col + ", lig = " + lig  + " val =" + val);
		}

		indicesColonne[col][lig] = val;
	}


	/**
	 * Indique si la partie est gagnée.
	 * @return
	 *        true si la partie est gagnée, false sinon.
	 */
	public boolean estGagne() {
		if (picross != null) {
			return picross.estGagne();
		} else {
			return gagne;
		}
	}


	/**
	 * Indique si la connexion au serveur est perdue.
	 * @return
	 *        true connexion est perdu, false sinon.
	 */
	public boolean connexionPerdue() {
		if (picross != null) {
			return false;
		} else {
			return connexionPerdue;
		}
	}

	/**
	 * Fixe le statut de connexion perdue.
	 * @param connexionPerdue
	 *        true si la connexion est perdue, false sinon.
	 */
	public  void setConnexionPerdue(final boolean connexionPerdue) {
		this.setLastChange(System.currentTimeMillis());
		this.connexionPerdue = connexionPerdue;
	}


	/**
	 * Change le statut gagne. Est utilisé par le client pour passer ce statut à true
	 * lors de la réception du message GAGNE.
	 * @param gagne
	 *        la valeur du statut.
	 */
	public void setGagne(final boolean gagne) {
		if (LOG_ON && COMM.isEnabledFor(Level.TRACE)) {
			COMM.trace("[PicrossApi::setGagne] gagne = " + gagne);
		}
		this.setLastChange(System.currentTimeMillis());
		this.gagne = gagne;
	}


	/**
	 * Retourne la hauteur de la grille de jeu. Bloque le thread appellant tant que
	 * les paramètres du jeu ne sont pas totalement initialisés. Le notify() est 
	 * envoyé dans la méthode indiquerInitialisationFinie().
	 * @return
	 *        la hauteur de la grille de jeu.
	 */
	public synchronized int getHauteur() {
		while (!initialisationFinie) {
			try {
				wait();
			} catch (InterruptedException e) {
				// on reboucle.
			}
		}

		return hauteur;
	}

	/**
	 * Retourne la largeur de la grille de jeu. Bloque le thread appellant tant que
	 * les paramètres du jeu ne sont pas totalement initialisés. Le notify() est 
	 * envoyé dans la méthode indiquerInitialisationFinie().
	 * @return
	 *        la largeur de la grille de jeu.
	 */
	public synchronized int getLargeur() {
		while (!initialisationFinie) {
			try {
				wait();
			} catch (InterruptedException e) {
				// on reboucle.
			}
		}
		return largeur;
	}

	/**
	 * Retourne la hauteur de la grille des indices sur les lignes.
	 * Bloque le thread appellant tant que les paramètres du jeu ne sont pas totalement
	 * initialisés. Le notify() est envoyé dans la méthode indiquerInitialisationFinie().
	 * @return
	 *        la hauteur de la grille des indices sur les lignes.
	 */
	public synchronized int getHauteurIndices() {
		while (!initialisationFinie) {
			try {
				wait();
			} catch (InterruptedException e) {
				// on reboucle.
			}
		}
		return hauteurIndices;
	}

	/**
	 * Retourne la largeur de la grille des indices sur les colonnes.
	 * Bloque le thread appellant tant que les paramètres du jeu ne sont pas totalement
	 * initialisés. Le notify() est envoyé dans la méthode indiquerInitialisationFinie().
	 * @return
	 *        la largeur de la grille des indices sur les colonnes.
	 */
	public synchronized int getLargeurIndices() {
		while (!initialisationFinie) {
			try {
				wait();
			} catch (InterruptedException e) {
				// on reboucle.
			}
		}
		return largeurIndices;
	}

	/**
	 * Indique que les paramètres du jeu sont tous initialisés et notifie
	 * le thead de l'application graphique pour le débloquer du wait().
	 */
	public synchronized void indiquerInitialisationFinie() {
		this.initialisationFinie = true;
		notify();
	}

	/**
	 * Retourne une valeur de la grille de jeu.
	 * @param col
	 *          colonne de la case.
	 * @param lig
	 *          ligne de la case.
	 * @return
	 *        valeur de la case.
	 */
	public int getValeur(final int col, final int lig) {
		return grille[col][lig];
	}

	/**
	 * Retourne une valeur de la grille des indices des lignes.
	 * @param col
	 *          colonne de la case.
	 * @param lig
	 *          ligne de la case.
	 * @return
	 *           valeur de la case.
	 */
	public int getIndicesLigne(final int col, final int lig) {
		return indicesLigne[col][lig];
	}

	/**
	 * Retourne une valeur de la grille des indices des colonnes.
	 * @param col
	 *          colonne de la case.
	 * @param lig
	 *          ligne de la case.
	 * @return
	 *           valeur de la case.
	 */
	public int getIndicesColonne(final int col, final int lig) {
		return indicesColonne[col][lig];
	}

	/**
	 * getter sur la propriété lastChange. 
	 * @return
	 *        la valeur de la propriété lastChange.
	 */
	public Long getLastChange() {
		return lastChange;
	}

	/**
	 * Change le proprité lastChange.
	 * @param lastChange
	 *        nouvelle valeur pour la date du dernier changement.
	 */
	public void setLastChange(final Long lastChange) {
		changeSupport.firePropertyChange("lastChange", this.lastChange, lastChange);
		this.lastChange = lastChange;
	}

	/**
	 * Ajoute un listerner en écoute sur la proprieté lastChange.
	 * @param listener
	 *        réference de l'instance à notifier lors des changements de la propriété lastChange.
	 */
	public void addPropertyChangeListener(final PropertyChangeListener listener) {
		changeSupport.addPropertyChangeListener(listener);
	}
}
